# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
import inspect
import functools
from enum import Enum

from agml.framework import AgMLSerializable
from agml.backend.tftorch import (
    get_backend, set_backend, user_changed_backend, StrictBackendError
)
from agml.data.managers.transform_helpers import (
    AlbumentationsTransformSingle,
    AlbumentationsTransformMask,
    AlbumentationsTransformCOCO,
    SameStateImageMaskTransform,
    NormalizationTransformBase,
    ScaleTransform,
    NormalizationTransform,
    OneHotLabelTransform,
    MaskToChannelBasisTransform
)
from agml.utils.logging import log


class TransformKind(Enum):
    Transform = 'transform'
    TargetTransform = 'target_transform'
    DualTransform = 'dual_transform'


# Shorthand form of the enum for testing explicit values.
t_ = TransformKind


class TransformManager(AgMLSerializable):
    """Manages the validation and application of transformations.

    This class serves as a helper for the `DataManager` class that
    focuses solely on applying transforms to the input images and
    annotations.

    There are three main transforms stored internally.

    1. `transform`: A transform which is applied to only the input image.
    2. `target_transform`: A transform which is applied only to the output.
    3. `dual_transform`: A transform which is applied to both.

    Note that the values of these transforms don't always directly match
    the provided value to the corresponding argument in the method
    `AgMLDataLoader.transform()`. An example is an albumentations
    transform, which, while being passed to `transform`, ends up being
    stored as a `dual_transform` internally. The above case applies to
    semantic segmentation and object detection transformations, but
    there is no `dual_transform` for image classification.
    """
    serializable = frozenset(
        ('task', 'transforms', 'time_inserted_transforms'))

    def __init__(self, task):
        self._task = task
        self._transforms = dict()

        # A user might want to apply a certain set of transforms first,
        # which are then followed by a different set of transforms. E.g.,
        # calling `loader.transform()` with one set of transforms and then
        # following with another call to `loader.transform()` with the
        # intention that the transforms in the second call will only be
        # applied after all of the transforms in the first call are.
        #
        # So, while the transform types are tracked in the `_transforms`
        # attribute, we track the moment that transforms are inserted
        # using this attribute, which is a list of different transforms.
        # The `apply()` method loops sequentially through each transform
        # in this list and applies them as required.
        self._time_inserted_transforms = []

    def get_transform_states(self):
        """Returns a copy of the existing transforms."""
        transform_dict = {}
        for name, state in self._transforms.items():
            transform_dict[name] = state.copy()
        return transform_dict

    def _pop_transform(self, t_type, search_param):
        """Removes a certain type of transform from the manager."""
        for i, tfm in enumerate(self._transforms[search_param]):
            if isinstance(tfm, t_type):
                self._transforms[search_param].pop(i)
                break
        for i, tfm in enumerate(self._time_inserted_transforms):
            if isinstance(tfm[1], t_type):
                self._time_inserted_transforms.pop(i)
                break

    def assign(self, kind, transform):
        """Assigns a new transform to the manager."""
        # Determine if the transform is being reset or unchanged.
        if transform == 'reset':
            self._transforms.pop(kind, None)
            new_time_transforms = []
            for tfm in self._time_inserted_transforms:
                if tfm[0] == kind:
                    continue
                new_time_transforms.append(tfm)
            self._time_inserted_transforms = new_time_transforms.copy()
            return
        elif transform is None:
            return

        # If an `albumentations` transform is passed to the `transform`
        # argument, then it is checked first to see if it is potentially
        # just processing the input image, and then passed as just a 
        # `transform` (or else it will clash in object detection tasks).
        # Otherwise, it is stored internally as a `dual_transform`.
        if transform is not None:
            try:
                if 'albumentations' in transform.__module__:
                    if t_(kind) == TransformKind.Transform:
                        if len(transform.processors) != 0:
                            kind = 'dual_transform'
                    if self._task == 'semantic_segmentation':
                        kind = 'dual_transform'
            except AttributeError:
                # Some type of object that doesn't have `__module__`.
                pass

        # We can only do this after the albumentations check, to ensure
        # that we are adding the transforms to the correct location.
        prev = self._transforms.get(kind, None)

        # Validate the transformation based on the task and kind.
        if self._task == 'image_classification':
            if t_(kind) == TransformKind.Transform:
                transform = self._maybe_normalization_or_regular_transform(transform)
            elif t_(kind) == TransformKind.TargetTransform:
                if isinstance(transform, tuple): # a special convenience case
                    if transform[0] == 'one_hot':
                        if transform[2] is not True: # removing the transform
                            self._pop_transform(OneHotLabelTransform, kind)
                            return
                        transform = OneHotLabelTransform(transform[1])
            else:
                raise ValueError("There is no `dual_transform` for image "
                                 "classification tasks. Please pass the "
                                 "input as a `transform` or `target_transform`.")
        elif self._task == 'image_regression':
            if t_(kind) == TransformKind.Transform:
                transform = self._maybe_normalization_or_regular_transform(transform)
            elif t_(kind) == TransformKind.TargetTransform:
                if isinstance(transform, tuple): # a special convenience case
                    if transform[0] == 'one_hot':
                        if transform[2] is not True: # removing the transform
                            self._pop_transform(OneHotLabelTransform, kind)
                            return
                        transform = OneHotLabelTransform(transform[1])
            else:
                pass
        elif self._task == 'semantic_segmentation':
            if t_(kind) == TransformKind.Transform:
                transform = self._maybe_normalization_or_regular_transform(transform)
            elif t_(kind) == TransformKind.TargetTransform:
                if isinstance(transform, tuple): # a special convenience case
                    if transform[0] == 'channel_basis':
                        if transform[2] is not True: # removing the transform
                            self._pop_transform(MaskToChannelBasisTransform, kind)
                        transform = MaskToChannelBasisTransform(transform[1])
                else:
                    transform = self._maybe_normalization_or_regular_transform(transform)
            else:
                transform = self._construct_image_and_mask_transform(transform)
        elif self._task == 'object_detection':
            if t_(kind) == TransformKind.Transform:
                transform = self._maybe_normalization_or_regular_transform(transform)
            elif t_(kind) == TransformKind.TargetTransform:
                pass
            else:
                transform = self._construct_image_and_coco_transform(transform)

        # Add the transformation to the internal storage.
        if transform is not None:
            if prev is not None:
                self._transforms[kind].append(transform)
            else:
                self._transforms[kind] = [transform]
            self._time_inserted_transforms.append((kind, transform))

    def apply(self, contents):
        """Applies a transform to a set of input data.

        This method controls the application of the actual transforms. It
        does this inside of a context that can control the application
        of the transform and manage errors more effectively.

        The method of application is mostly similar across the different
        tasks, as the differentiation of the methods is taken care of
        in the `assign` method and each transform is converted to act
        like a generic method with simple input arguments.

        The hierarchy in which transforms are applied is:


             transform  ->  --------|
                                    |----->   dual_transform
             target_transform ->  --|

        Furthermore, image resizing takes place before any transformations
        are applied. After the transforms are applied in this order, they
        returned and if passed again, they will have a different transform
        applied to them. The state is independent of the images passed.
        """
        image, annotation = contents

        # Iterate through the different transforms.
        for (kind, transform) in self._time_inserted_transforms:
            if t_(kind) == TransformKind.Transform:
                image = self._apply_to_objects(
                    transform, (image, ), kind)
            if t_(kind) == TransformKind.TargetTransform:
                annotation = self._apply_to_objects(
                    transform, (annotation, ), kind)
            if t_(kind) == TransformKind.DualTransform:
                image, annotation = self._apply_to_objects(
                    transform, (image, annotation), kind)

        # Return the processed image and annotation.
        return image, annotation

    @staticmethod
    def _apply_to_objects(transform, contents, kind):
        """Applies the actual transformations in a context."""
        try:
            return transform(*contents)
        except Exception as e:
            default_msg = (f"Encountered an error when attempting to apply "
                           f"a transform ({transform}) of kind '{kind}' to "
                           f"objects: {contents}. See the above traceback.")

            # A specific case of exception where the image first needs
            # to be converted to a PIL image before being used in a
            # general `torchvision.transforms` pipeline.
            if "PIL" in str(e):
                raise TypeError("If using a `torchvision.transforms` pipeline "
                                "when not in PyTorch training mode, you need "
                                "to include `ToTensor()` in the pipeline.")

            # Otherwise, raise the default exception.
            raise Exception(default_msg)

    def _maybe_normalization_or_regular_transform(self, transform):
        """Dispatches to the correct single-image transform construction."""
        if isinstance(transform, tuple):
            if transform[0] == 'normalize':
                return self._build_normalization_transform(transform)
        return self._construct_single_image_transform(transform)

    def _build_normalization_transform(self, transform):
        """Constructs a normalization transform if passed.

        This is a special case for transforms passed by the `normalize_images`
        'method of the `AgMLDataLoader`, since these are treated almost as
        their own independent management system in terms of resetting or
        applying them in a different method. This is called by `assign`.
        """
        # First, we check if a normalization transform already exists
        # within the transform dict, and then we get its location.
        norm_transform_index, norm_transform_index_time = -1, -1
        try:
            for i, t in enumerate(self._transforms['transform']):
                if isinstance(t, NormalizationTransformBase):
                    norm_transform_index = i
                    break
            for i, (_, t) in enumerate(self._time_inserted_transforms):
                if isinstance(t, NormalizationTransformBase):
                    norm_transform_index_time = i
                    break
        except:
            self._transforms['transform'] = []

        if transform[1] == 'scale':
            tfm = ScaleTransform(None)
            if norm_transform_index != -1:
                self._transforms['transform'][norm_transform_index] = tfm
                self._time_inserted_transforms[norm_transform_index_time] \
                    = ('transform', tfm)
            else:
                self._transforms['transform'].append(tfm)
                self._time_inserted_transforms.append(('transform', tfm))
        elif hasattr(transform[1], 'mean') or transform[1] == 'imagenet':
            try:
                mean, std = transform[1].mean, transform[1].std
            except AttributeError:
                # Default ImageNet mean and std.
                mean = [0.485, 0.456, 0.406]
                std = [0.229, 0.224, 0.225]
            tfm = NormalizationTransform((mean, std))
            if norm_transform_index != -1:
                self._transforms['transform'][norm_transform_index] = tfm
                self._time_inserted_transforms[norm_transform_index_time] \
                    = ('transform', tfm)
            else:
                self._transforms['transform'].append(tfm)
                self._time_inserted_transforms.append(('transform', tfm))
        elif transform[1] == 'reset':
            if norm_transform_index != -1:
                self._transforms['transform'].pop(norm_transform_index)
                self._time_inserted_transforms.pop(norm_transform_index_time)
        return None

    # The following methods implement different checks which validate
    # as well as process input transformations, and manage the backend.
    # The transforms here will be also checked to match a specific
    # backend. Alternatively, the backend will dynamically be switched.

    @staticmethod
    def _construct_single_image_transform(transform):
        """Validates a transform which is applied to a single image.

        This is used for image classification transforms, which only
        apply to the input image, as well as other general tasks whenever
        a transform that is only applied to the image is passed, e.g.,
        a visual augmentation like random contrasting.
        """
        # This case is used for clearing a transformation.
        if transform is None:
            return None

        # A general functional transformation. We don't verify what happens in
        # the function. The only check which occurs here that the signature of
        # the function is valid. Note that this also includes partial methods.
        elif isinstance(transform, (types.FunctionType, functools.partial)):
            sig = inspect.signature(transform).parameters
            if not len(sig) == 1:
                raise TypeError("Expected a single-image transform passed "
                                "to `transform` to accept one input image, "
                                f"instead got {len(sig)} parameters.")
            return transform

        # An `albumentations` transform to be applied to the image. This
        # wraps the transform into a method which treats it as a regular
        # functional transform, e.g. no keyword arguments (for easy use).
        elif 'albumentations' in transform.__module__:
            return AlbumentationsTransformSingle(transform)

        # A set of `torchvision` transforms wrapped into a `T.Compose` object
        # or just a single transformation. This simply confirms the backend.
        elif 'torchvision' in transform.__module__:
            if get_backend() != 'torch':
                if user_changed_backend():
                    raise StrictBackendError(change = 'tf', obj = transform)
                set_backend('tf')
            return transform

        # A `tf.keras.Sequential` preprocessing model or an individual
        # Keras preprocessing layer. This simply confirms the backend.
        elif 'keras' in transform.__module__:
            if get_backend() != 'tf':
                if user_changed_backend():
                    raise StrictBackendError(change = 'torch', obj = transform)
                set_backend('torch')
            return transform

        # Otherwise, it may be a transform from a (lesser-known) third-party
        # library, in which case we just return it as a callable. Transforms
        # which are used in a more complex manner should be passed as decorators.
        return transform

    @staticmethod
    def _construct_image_and_mask_transform(transform):
        """Validates a transform for an image and annotation mask.

        This is used for a semantic segmentation transform. Such
        transformations should be passed as the following:

        - An `albumentations` transform pipeline that may include
          spatial and/or visual augmentation.
        - A method to independently or dually apply transformations
          to the image and annotation mask.
        - A `torchvision.transforms` or `tf.keras.Sequential` pipeline
          which will be applied to the image and mask using the same
          random seed, for reproducibility. Use the provided method
          `generate_keras_segmentation_dual_transform` for this.
        """
        # This case is used for clearing a transformation.
        if transform is None:
            return None

        # A general functional transformation. We don't verify what happens in
        # the function. The only check which occurs here that the signature of
        # the function is valid. Note that this also includes partial methods.
        elif isinstance(transform, (types.FunctionType, functools.partial)):
            sig = inspect.signature(transform).parameters
            if not len(sig) == 2:
                raise TypeError(f"Expected a semantic segmentation transform "
                                f"passed to `transform` to accept two args: "
                                f"an input image and an annotation mask, "
                                f"instead got {len(sig)} parameters.")
            return transform

        # An `albumentations` transform to be applied to the image. This
        # wraps the transform into a method which treats it as a regular
        # functional transform, e.g. no keyword arguments (for easy use).
        elif 'albumentations' in transform.__module__:
            return AlbumentationsTransformMask(transform)

        # If we have the case of a transform that needs to be applied to
        # both the input and the output mask simultaneously, then we wrap
        # that into a class which undertakes that behavior. This happens
        # when the signature of the input function accepts only one input
        # parameter or it belongs to `torchvision` transform (not Keras).
        if len(inspect.signature(transform).parameters) == 1:
            if 'torchvision' in transform.__module__:
                if get_backend() != 'torch':
                    if user_changed_backend():
                        raise StrictBackendError(
                            change = 'tf', obj = transform)
                    set_backend('tf')
            elif 'keras.layers' in transform.__module__:
                if get_backend() != 'tf':
                    if user_changed_backend():
                        raise StrictBackendError(
                            change = 'torch', obj = transform)
                    set_backend('torch')
                log('Got a Keras transformation for a dual image and '
                    'mask transform. If you are passing preprocessing '
                    'layers to this method, then use `agml.data.experimental'
                    '.generate_keras_segmentation_dual_transform` in order '
                    'for the random state to be applied properly.', 'warning')
            return SameStateImageMaskTransform(transform)

        # Another type of transform, most likely some form of transform
        # class. No checks are applied here, since we can't account for
        # each of the potential cases of the transformations.
        return transform

    @staticmethod
    def _construct_image_and_coco_transform(transform):
        """Validates a transform for an image and COCO JSON dictionary.

        This is used for object detection transforms. Such transformations
        should be wrapped into a method (unless they are albumentations
        transforms). The method should accept two input arguments, the image
        and the COCO JSON dictionary, and return the two respectively.
        """
        # This case is used for clearing a transformation.
        if transform is None:
            return None

        # A general functional transformation. We don't verify what happens in
        # the function. The only check which occurs here that the signature of
        # the function is valid. Note that this also includes partial methods.
        elif isinstance(transform, (types.FunctionType, functools.partial)):
            sig = inspect.signature(transform).parameters
            if not len(sig) == 2:
                raise TypeError(f"Expected a object detection transform passed "
                                f"to `transform` to accept two args: an input "
                                f"image and a COCO JSON dictionary, instead "
                                f"got {len(sig)} parameters.")
            return transform

        # An `albumentations` transform to be applied to the image. This
        # wraps the transform into a method which treats it as a regular
        # functional transform, e.g. no keyword arguments (for easy use).
        elif 'albumentations' in transform.__module__:
            return AlbumentationsTransformCOCO(transform)

        # Another type of transform, most likely some form of transform
        # class. No checks are applied here, since we can't account for
        # each of the potential cases of the transformations.
        return transform





# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Experimental data utilities that are in development.
"""
import functools

import numpy as np


__all__ = ['generate_keras_segmentation_dual_transform']


def generate_keras_segmentation_dual_transform(*layers):
    """Generates a `dual_transform` pipeline from Keras preprocessing layers.

    This method takes in Keras preprocessing layers and generates a
    transformation pipeline for the `dual_transform` argument in
    *semantic segmentation* loaders, which applies the transform in the
    same fashion to both the image and annotation.

    This is due to the fact that TensorFlow has its operation-level
    random states different than its module-level random state, so
    the layers need to have their seeds manually set in order to work.

    In essence, for each of the preprocessing layers passed, this
    method conducts the following operations:

    > def preprocessing_transform(image, annotation):
    >    layer = functools.partial(KerasPreprocessingLayer, **kwargs)
    >    seed = np.random.randint(BUFFER_SIZE) # up to sys.maxsize
    >    image = layer(image, seed = seed)
    >    annotation = layer(annotation, seed = seed)
    >    return image, annotation

    It then repeats this transform for all of the preprocessing layers
    passed, and returns a method which has this behavior wrapped into
    it and can perform it when the preprocessing is actually conducted.

    Parameters
    ----------
    layers : Any
       Either a Sequential model with preprocessing layers, or a
       set of instantiated preprocessing layers.

    Returns
    -------
    """
    import tensorflow as tf
    if len(layers) == 1:
        if isinstance(layers[0], tf.keras.Sequential):
            layers = layers[0].layers

    # These methods perform the behavior indicated in the
    # code snippet above (for each of the layers given).
    def _single_preprocessing_layer_base(layer_, build_dict):
        def _internal(image, annotation, seed):
            instantiated_layer = functools.partial(layer_, **build_dict)
            seed_update = {}
            if seed is not None:
                seed_update['seed'] = seed
            image = instantiated_layer(**seed_update)(image)
            annotation = instantiated_layer(**seed_update)(annotation)
            return image, annotation
        return _internal

    preprocessing_methods, use_seeds = [], []
    for layer in layers:
        config = layer.get_config()
        if 'seed' in config:
            config.pop('seed')
            use_seeds.append(True)
        else:
            use_seeds.append(False)
        preprocessing_methods.append(
            _single_preprocessing_layer_base(layer.__class__, config))

    def _execute_preprocessing(layers_, use_seeds_):
        def _execute(image, annotation):
            for p_layer, seed_ in zip(layers_, use_seeds_):
                seed = np.random.randint(2147483647) if seed_ else None
                image, annotation = p_layer(image, annotation, seed = seed)
            return image, annotation
        return _execute
    return _execute_preprocessing(preprocessing_methods, use_seeds)










# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import io
import sys
import json
import datetime
import subprocess as sp
from dataclasses import dataclass
import xml.etree.ElementTree as ET
from collections.abc import Sequence
from typing import List, Union

from dict2xml import dict2xml

from agml.framework import AgMLSerializable
from agml.backend.config import synthetic_data_save_path
from agml.synthetic.options import HeliosOptions
from agml.synthetic.options import AnnotationType, SimulationType
from agml.synthetic.config import load_default_helios_configuration
from agml.synthetic.compilation import (
    HELIOS_BUILD, HELIOS_EXECUTABLE, XML_PATH, PROJECT_PATH
)
from agml.synthetic.converter import HeliosDataFormatConverter


@dataclass
class GenerationInstanceOptions:
    canopy: str
    num_images: int
    annotation_type: Union[AnnotationType, str]
    simulation_type: Union[SimulationType, str]
    labels: List[str]
    output_dir: str

    def __post_init__(self):
        if self.num_images < 1:
            raise ValueError("The number of images cannot be negative.")

        try:
            self.annotation_type = AnnotationType(self.annotation_type)
        except ValueError:
            raise ValueError(
                "Expected either an `AnnotationType` parameter "
                "or corresponding string for 'annotation_type'. "
                "Valid annotation types: `object_detection`, "
                "`semantic_segmentation`, `instance_segmentation`, "
                "or for no annotations at all, `none`.")

        try:
            self.simulation_type = SimulationType(self.simulation_type)
        except ValueError:
            raise ValueError(
                "Expected either a `SimulationType` parameter "
                "or corresponding string for 'simulation_type'. "
                "Valid simulation types: rgb`, `lidar`.")

        valid_labels = ['trunks', 'leaves', 'fruits', 'branches']
        if not all(i in valid_labels for i in self.labels):
            raise ValueError(f"Got one or more invalid labels: {self.labels}."
                             f"Valid labels: {[*valid_labels]}.")
        if self.canopy in ['Tomato', 'Strawberry'] and 'trunks' in self.labels:
            raise ValueError("Tomato and Strawberry canopies do not have any trunks, "
                             "so trying to use the `trunks` label is impossible.")

    @classmethod
    def _empty(cls):
        return cls.__new__(cls)


class HeliosDataGenerator(AgMLSerializable):
    """Generates a synthetic agricultural dataset using Helios.

    The `HeliosDataGenerator` provides a Python interface to the Helios C++
    data generation software (https://github.com/PlantSimulationLab/Helios),
    used for generating synthetic agricultural data and annotations. It can
    be used to generate a wide variety of synthetic datasets, with vast
    customizability through the  available options and parameters, including
    both RGB and LiDAR simulations, annotations from object detection to
    semantic segmentation, and a variety of different crops.

    Parameters
    ----------
    options : HeliosOptions
        A `HeliosOptions` instance containing pre-initialized parameters.
        If you want to directly use the pre-set parameters, then you can
        also use the keyword argument `canopy` to bypass the options.
    canopy : str
        A specific canopy to generate images for. This can be passed for
        in place of `options` to initialize the default options.
    """

    def __new__(cls, *args, **kwargs):
        # Run the configuration check. Notice that we run this here because
        # users may not necessarily want to use Helios on first installing
        # AgML, but the `synthetic` module is loaded into the main AgML
        # module as in the top-level `__init__.py` file. By putting the check
        # here, it ensures that Helios is only installed if and when data is
        # actually being generated, and not just by a standard import.
        from .config import _check_helios_installation
        _check_helios_installation()
        del _check_helios_installation

        # Return a `HeliosDataGenerator`.
        return super(HeliosDataGenerator, cls).__new__(cls)

    def __init__(self, options: HeliosOptions = None, *, canopy = None):
        if options is None and canopy is not None:
            self._options = HeliosOptions(canopy)
        elif options is not None and canopy is None:
            self._options = options
        else:
            raise ValueError(
                "The `HeliosDataGenerator` needs either "
                "a set of `HeliosOptions` or a `canopy`.")
        self._canopy = self._options._canopy

        # Rather than constantly passing the different parameters used during
        # generation, such as the annotation type or the output directory,
        # throughout each method, whenever data generation begins, this class
        # is instantiated with all of the parameters for convenience. This
        # also enables wrapping all of the checks into a single method which is
        # contained within the external class as opposed to this class.
        self._generation_options = GenerationInstanceOptions._empty()

    @property
    def options(self) -> HeliosOptions:
        return self._options

    @options.setter
    def options(self, value: HeliosOptions):
        self._canopy = value._canopy
        self._options = value

    @staticmethod
    def _sanitize_name(name):
        """Cleans a name to represent a valid folder destination."""
        return name.replace(' ', '-').replace('~', '-')

    def _convert_options_to_xml(self):
        """Converts the `HeliosOptions` parameters to XML format."""
        # Convert all of the parameters to XML format.
        tree = ET.parse(io.StringIO(dict2xml(
            {'helios': self._prepare_parameters_for_generation()})))

        # Update the camera tags.
        root = tree.getroot()
        for child in root:
            if child.tag == 'camera_position':
                child.tag = 'globaldata_vec3'
                child.set('label', 'camera_position')
            if child.tag == 'camera_lookat':
                child.tag = 'globaldata_vec3'
                child.set('label', 'camera_lookat')
            if child.tag == 'image_resolution':
                child.tag = 'globaldata_int2'
                child.set('label', 'image_resolution')

        # Return the modified tree.
        return tree

    def _prepare_parameters_for_generation(self):
        """Prepares the provided Helios parameter options for generation."""
        parameters = self._options._to_dict()

        # Divide the image resolution in half. For some reason, Helios
        # appears to generate images with twice the provided resolution,
        # so in order to fix this, we need to divide the resolution in
        # two here so that it is correct during generation.
        parameters['camera']['image_resolution'] = \
            [i // 2 for i in parameters['camera']['image_resolution']]

        # Generate the ground parameters.
        ground_params = self._load_ground_parameters()
        parameters['Ground'] = ground_params

        # Convert the canopy parameters to the Helios string format.
        self._convert_dict_params_to_string(parameters)

        # Create a fresh dictionary and put all of the parameters in it.
        canopy_parameters = {
            self._canopy + "Parameters": parameters['canopy'],
            'Ground': parameters['Ground']}
        xml_params = {'canopygenerator': canopy_parameters}
        if self._generation_options.simulation_type == 'lidar':
            xml_params['scan'] = parameters['lidar']
        else:
            xml_params[''] = parameters['camera']

        # Return the parsed dictionary.
        return xml_params

    @staticmethod
    def _convert_dict_params_to_string(d):
        """Converts dictionary int/float parameters to strings."""
        for key, value in d.items():
            for param, param_value in value.items():
                if isinstance(param_value, Sequence) \
                        and not isinstance(param_value, str):
                    if isinstance(param_value[0], Sequence):
                        param_value = [str(a).replace(",", "").replace(
                            "[", " ").replace("]", "") for a in param_value]
                        value[param] = \
                            f' {" ".join(param_value)} '
                    else:
                        value[param] = \
                            f' {" ".join([str(a) for a in param_value])} '
                else:
                    value[param] = f' {str(param_value)} '

    @staticmethod
    def _load_ground_parameters():
        """Loads the parameters for the `Ground` plant type."""
        cfg = load_default_helios_configuration()
        return cfg['canopy']['parameters']['Ground']

    @staticmethod
    def _parse_output_path(output_path: str):
        output_path = os.path.abspath(output_path)
        if os.path.exists(output_path) and os.path.isdir(output_path):
            if any(os.scandir(output_path)):
                raise OSError(f"The provided directory for generation, "
                              f"'{output_path}', is not empty. Please "
                              f"clear out the contents of this directory "
                              f"and try again.")
        else:
            os.makedirs(output_path)

    def _write_config(self, *, cfg_file, xml_file):
        """Writes the config file that Helios loads from."""
        # Construct the string with the output.
        xml_file = os.path.realpath(xml_file)
        cfg = f"{self._generation_options.num_images}\n" \
              f"{self._generation_options.annotation_type.value}\n" \
              f"{self._generation_options.simulation_type.value}\n" \
              f"{' '.join(self._generation_options.labels)}\n" \
              f"{xml_file}\n{self._generation_options.output_dir}"

        # Write the configuration file to its output path.
        with open(os.path.join(PROJECT_PATH, cfg_file), 'w') as f:
            f.write(cfg)

    def generate(self, *,
                 name: str = None,
                 num_images: int = 5,
                 output_dir: os.PathLike = None,
                 convert_data: bool = True):
        """Generates a batch of synthetic data using Helios.

        This method is the one which conducts the actual data generation using
        Helios. By this point, the `HeliosDataGenerator` should have an initialized
        `HeliosOptions` class which contains all of the scene and canopy generation
        parameters, set to the user preference, and this method takes in the name
        of the run itself as well as the amount of images to generate, the final
        parameters which are used to generate the actual data.

        The `name` parameter here refers to an optional name which can be assigned
        to customize the dataset that is being generated (e.g., if it is a dataset
        of 100 strawberry plants, then it could be 'strawberry100'). If no specific
        name is provided, then the generator defaults to using today's date and time
        (e.g., a run at 12:15 PM on January 12, 2024 would be 'helios-01122024-1215').

        If multiple camera views are provided as part of the `HeliosOptions`, then the
        number of images will actually end up being `num_images` times the number of
        camera views, e.g., if there are 5 camera views and 50 images to be generated,
        then there will be 250 total images in the resulting dataset. The final output
        directory structure ends up being something like this:

        generated_images
            image0
                view00000
                view00001
                ...
            image1
                view00000
                view00001
                ...
            ...

        NOte that this is only the post-Helios output structure. By default, after
        the dataset is generated, the annotations will be converted from the Helios
        format to the relevant format within AgML. For example, annotation text files
        will be converted to COCO JSON, and the directory structure of the dataset
        will also be changed accordingly. If you want to maintain the integrity of
        the generated data for Helios, then you can provide the optional parameter
        `convert_data = False`, which will leave data as generated by Helios.

        Regardless of whether annotations are converted to the AgML format from the
        Helios format, the output directory will contain an extra hidden directory
        `.metadata`, which will contain a few extra files:

            1. The XML file which contains the parameters in the `HeliosOptions`.
            2. The `config.txt` file which is used for generation.
            3. An extra `meta.json` file which contains some additional information,
               such as the classes in the dataset, and the annotation type.

        These are used by the `AgMLDataLoader` when loading a custom Helios-generated
        dataset, and can also be used to recreate a dataset from its original parameters.

        Parameters
        ----------
        name : str
            The name of the run (and thus the generated dataset).
        num_images : int
            The number of images to generate in the dataset.
        output_dir : str
            A custom output directory where data will be generated to. If it is
            a path such as `/root/path`, then the data will be generated to the
            path `/root/path/name`. If `/root/path/name` is passed, then it will
            be saved directly to there (not `/root/path/name/name`).
        convert_data : {bool, optional}
            Whether to convert data to the AgML format after generation. If left to
            False, then data is left in the corresponding Helios format.

        Notes
        -----
        Unless a custom path is passed to this method, all datasets will be saved to
        the default *synthetic* data generation path. Note that this is distinct from
        `agml.backend.data_save_path()`. This defaults to `~/.agml/synthetic`, but can
        be changed using `agml.backend.set_synthetic_save_path()`, and then accessed
        using `agml.backend.synthetic_data_save_path()`. The distinction is important.
        """
        # Construct and sanitize the name of the run.
        if name is not None:
            name = self._sanitize_name(name)
        else:
            date = datetime.datetime.now().strftime('%m%d%Y-%H%M')
            name = f'helios-{date}'

        # Construct the output path of the dataset.
        if output_dir is not None:
            if os.path.basename(output_dir) == name:
                output_dir = os.path.dirname(output_dir)
        else:
            output_dir = synthetic_data_save_path()
        output_dir = os.path.join(output_dir, name)
        if os.path.exists(output_dir):
            self._parse_output_path(output_dir)
        else:
            os.makedirs(output_dir)

        # Construct the `GenerationInstanceOptions` class from the `HeliosOptions`
        # parameters and the newly passed input parameters.
        self._generation_options = GenerationInstanceOptions(
            canopy = self._canopy,
            num_images = num_images,
            annotation_type = self._options.annotation_type,
            simulation_type = self._options.simulation_type,
            labels = self._options.labels,
            output_dir = output_dir)

        # Create the output metadata directory.
        metadata_dir = os.path.join(output_dir, '.metadata')
        os.makedirs(metadata_dir)

        # Construct the XML parameter style file.
        xml_options = self._convert_options_to_xml()
        xml_file_base = f"style_{name}.xml"
        xml_options.write(os.path.join(XML_PATH, xml_file_base))
        xml_options.write(os.path.join(metadata_dir, xml_file_base))

        # Write the actual configuration file.
        cfg_file = os.path.join(PROJECT_PATH, f'config_{name}.txt')
        self._write_config(
            cfg_file = cfg_file, xml_file = os.path.join(XML_PATH, xml_file_base))
        self._write_config(
            cfg_file = os.path.join(metadata_dir, f'config_{name}.txt'),
            xml_file = os.path.join(metadata_dir, xml_file_base))

        # Run the actual data generation with the executable.
        process = sp.Popen([HELIOS_EXECUTABLE, cfg_file], stdout = sp.PIPE,
                           stderr = sp.STDOUT, cwd = HELIOS_BUILD,
                           universal_newlines = True,)
        for line in iter(process.stdout.readline, ""):
            sys.stdout.write(line)
        process.stdout.close()
        process.wait()

        # Remove the configuration and style files.
        os.remove(cfg_file)
        os.remove(os.path.join(XML_PATH, xml_file_base))

        # Save the extra metadata.
        with open(os.path.join(metadata_dir, 'meta.json'), 'w') as f:
            json.dump({'labels': self._generation_options.labels,
                       'image_size': self._options.camera.image_resolution,
                       'generation_date': datetime.datetime.now().strftime(
                           '%B %d, %Y %H:%M')}, f, indent = 4)

        # Check that the process successfully completed.
        if process.returncode != 0:
            if process.returncode == -11:
                raise OSError(f"Encountered an error when generating synthetic "
                              f"data. Process returned code {process.returncode}, "
                              f"suggesting that the program ran out of memory. Try "
                              f"passing a smaller environment for generation.")
            else:
                raise OSError(f"Encountered an error when generating synthetic "
                              f"data. Process returned code {process.returncode}.")

        # Convert the dataset format.
        if convert_data:
            cvt = HeliosDataFormatConverter(output_dir)
            cvt.convert()




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn


@torch.jit.script
def accuracy(output, target):
    """Computes the accuracy between `output` and `target`."""
    batch_size = target.size(0)
    _, pred = torch.topk(output, 1, 1)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    correct_k = correct[:1].reshape(-1).float().sum(0, keepdim = True)
    return correct_k.mul_(100.0 / batch_size)


class Accuracy(nn.Module):
    """A metric to compute accuracy for image classification tasks.

    This class is used as a wrapper around accuracy calculations, which allows for
    accumulation of predictions over time. The `update` method can be used to update
    data, then `compute` to get the calculated accuracy, and finally `reset` can be
    used to reset the accumulators to an empty state, allowing new calculations.
    """

    def __init__(self):
        # Construct the data accumulators.
        super(Accuracy, self).__init__()
        self._prediction_data, self._truth_data = [], []

    def update(self, pred_data, gt_data):
        """Updates the state of the accuracy metric.

        The `pred_data` and `gt_data` arguments should both be sequences of integer
        labels (e.g., an array of one-dimension), *not* one-hot labels. So, any
        activation operations or softmax operations must be applied before inputting
        data into the accuracy metric.
        """
        if not len(pred_data) == len(gt_data):
            raise ValueError("Predictions and truths should be the same length.")
        self._prediction_data.extend(pred_data)
        self._truth_data.extend(gt_data)

    def compute(self):
        """Computes the accuracy between the predictions and ground truths."""
        return accuracy(torch.tensor(self._prediction_data),
                        torch.tensor(self._truth_data))

    def reset(self):
        """Resets the accumulator states."""
        del self._prediction_data, self._truth_data
        self._prediction_data, self._truth_data = [], []

